import tomllib
from pathlib import Path
from typing import Any

PYPROJ = "pyproject.toml"


def load_conf(path: Path) -> dict[str, Any]:
    with (path / PYPROJ).open("rb") as fp:
        return tomllib.load(fp)


def get_members() -> list[str]:
    root = get_workspace_root()
    root_conf = load_conf(root)
    members: list[str] = (
        root_conf.get("tool", {}).get("uv", {}).get("workspace", {}).get("members", [])  # pyright:ignore[reportAny]
    )
    return members


def get_dependencies(path: Path) -> tuple[list[str], list[str]]:
    conf = load_conf(path)
    all_deps: list[str] = conf["project"].get("dependencies", [])  # pyright:ignore[reportAny]
    try:
        sources: dict[str, dict[str, bool]] = conf["tool"]["uv"]["sources"]
    except KeyError as e:
        raise KeyError(f"No tool.uv.sources table for '{path}'") from e

    ext_deps: list[str] = []
    int_deps: list[str] = []
    for d in all_deps:
        if d in sources:
            if sources[d]["workspace"]:
                int_deps.append(d)
                continue
        ext_deps.append(d)
    return (ext_deps, int_deps)


def find_package_dir(name: str, members: list[str]) -> Path:
    root = get_workspace_root()
    for glob in members:
        packages = sorted(root.glob(glob))
        for p in packages:
            if p.name == name:
                return p.resolve()
    raise ValueError(f"Couldn't find package '{name}'")


def get_workspace_root() -> Path:
    root = _find_upwards(Path.cwd())
    if not root:
        raise ValueError("Didn't find the workspace root. Expected to find a .git directory.")
    return root


def _find_upwards(cwd: Path) -> Path | None:
    if cwd == Path(cwd.root) or cwd == cwd.parent:
        return None
    elif (cwd / ".git").exists():
        return cwd
    return _find_upwards(cwd.parent)



from pathlib import Path
from typing import Any

from hatchling.builders.config import BuilderConfig
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.builders.sdist import SdistBuilder
from hatchling.plugin import hookimpl

from hatch_una import util


class UnaBuildHook(BuildHookInterface[BuilderConfig]):
    """
    Force-include all needed internal monorepo dependencies.
    """

    PLUGIN_NAME = "una-build"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        via_sdist = Path("PKG-INFO").exists()
        if via_sdist:
            raise ValueError("Una doesn't work for wheels built from sdist")

        print("una-build: Injecting internal dependencies")

        # load the config for this package
        path = Path(self.root)
        root = util.get_workspace_root()
        root_conf = util.load_conf(root)
        members: list[str] = (
            root_conf.get("tool", {}).get("uv", {}).get("workspace", {}).get("members", [])  # pyright:ignore[reportAny]
        )
        _, int_deps = util.get_dependencies(path)

        if not int_deps:
            # this is fine, the package doesn't import anything internally
            return

        add_dep_files: dict[str, str] = {}
        package_dirs = [util.find_package_dir(d, members) for d in int_deps]
        for package_dir in package_dirs:
            finder = SdistBuilder(str(package_dir))
            files = [Path(f.path) for f in finder.recurse_selected_project_files()]
            for f in files:
                add_dep_files[str(f)] = str(f.relative_to(package_dir))

        build_data["force_include"] = {
            **build_data["force_include"],
            **add_dep_files,
        }


@hookimpl
def hatch_register_build_hook() -> type[UnaBuildHook]:
    return UnaBuildHook



from pathlib import Path
from typing import Any

from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl

from hatch_una import util


class UnaMetaHook(MetadataHookInterface):
    """
    Inject needed third-party dependencies into project.dependencies.
    """

    PLUGIN_NAME = "una-meta"

    def update(self, metadata: dict[str, Any]) -> None:
        print("una-meta: Injecting transitive external dependencies")

        path = Path(self.root)
        ext_deps, int_deps = util.get_dependencies(path)

        via_sdist = Path("PKG-INFO").exists()
        if via_sdist:
            raise ValueError("Una doesn't work for wheels built from sdist")

        members: list[str] = util.get_members()

        add_deps: list[str] = []
        for dep_name in int_deps:
            dep_path = util.find_package_dir(dep_name, members)

            # load all third-party dependencies from this internal dependency into the
            # project.dependencies table
            dep_deps, _ = util.get_dependencies(dep_path)
            dep_deps = [d.strip() for d in dep_deps]
            add_deps.extend(dep_deps)

        metadata["dependencies"] = list(set(ext_deps + add_deps))


@hookimpl
def hatch_register_metadata_hook() -> type[UnaMetaHook]:
    return UnaMetaHook
